use std::rc::Rc;

use malachite::{
    base::{num::basic::traits::Zero, rounding_modes::RoundingMode},
    rational::{Rational, conversion::primitive_float_from_rational::FloatConversionError},
};

use super::{Environment, Node, Precedence};

pub type ConstantValue = malachite::rational::Rational;

#[derive(Clone, Debug, PartialEq, PartialOrd)]
pub struct Constant {
    value: ConstantValue,
}

impl From<ConstantValue> for Constant {
    fn from(value: ConstantValue) -> Self {
        Self::new(value)
    }
}

impl Node for Constant {
    fn evaluate(&self, _: &mut super::Environment) -> Result<Rc<super::NodeEnum>, String> {
        Ok(Rc::new(self.clone().into()))
    }

    fn as_string(&self, _env: Option<&Environment>) -> String {
        if self.value == Rational::ZERO {
            return "0".to_owned();
        }

        if self.value.approx_log() < 50. {
            let x: Result<f64, FloatConversionError> = self.value.clone().try_into();

            match x {
                Ok(v) => {
                    return v.to_string();
                }
                Err(_) => {}
            }
        }

        if let Some((man, exp, _)) = self
            .value
            .sci_mantissa_and_exponent_round_ref::<f64>(RoundingMode::Nearest)
        {
            format!("{}*2^{}", man, exp)
        } else {
            self.value.to_string()
        }
    }

    fn precedence(&self) -> Precedence {
        Precedence::Primary
    }
}

impl Constant {
    pub fn new(value: ConstantValue) -> Self {
        Self { value }
    }

    pub fn get_value(&self) -> &ConstantValue {
        &self.value
    }

    pub fn set_value(&mut self, value: ConstantValue) {
        self.value = value;
    }
}
